Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Source file Interfaces.ml
12345678910111213141516171819202122232425262728293031323334353637383940414243444546474849505152535455565758596061626364656667686970717273747576777879808182838485868788(* This file is free software, part of Zipperposition. See file "license" for more details. *)(** {1 Interfaces} *)moduletypeEQ=sigtypetvalequal:t->t->boolendmoduletypeHASH=sigincludeEQvalhash:t->intendmoduletypeORD=sigtypetvalcompare:t->t->intend(** Print a type t *)moduletypePRINT=sigtypetvalpp:tCCFormat.printervalto_string:t->stringendmoduletypePRINT1=sigtype'atvalpp:'aCCFormat.printer->'atCCFormat.printervalto_string:'aCCFormat.printer->'at->stringendmoduletypePRINT2=sigtype('a,'b)tvalpp:'aCCFormat.printer->'bCCFormat.printer->('a,'b)tCCFormat.printervalto_string:'aCCFormat.printer->'bCCFormat.printer->('a,'b)t->stringendmoduletypePRINT3=sigtype('a,'b,'c)tvalpp:'aCCFormat.printer->'bCCFormat.printer->'cCCFormat.printer->('a,'b,'c)tCCFormat.printervalto_string:'aCCFormat.printer->'bCCFormat.printer->'cCCFormat.printer->('a,'b,'c)t->stringend(** Register printers by name *)moduletypePRINT_OVERLOAD=sigtypetvalpp_with:string->tCCFormat.printervaladd_printer:string->tCCFormat.printer->unitvalset_default_printer:string->unit(** Used by PRINT.pp... *)endmoduletypePRINT_DE_BRUIJN=sigtypettypetermtypeprint_hook=int->termCCFormat.printer->Format.formatter->term->bool(** User-provided hook that can be used to print terms (for composite cases)
before the default printing occurs. The int argument is the De Bruijn
depth in the term.
A hook takes as arguments the depth and the recursive printing function
that it can use to print subterms.
A hook should return [true] if it fired, [false] to fall back
on the default printing. *)valpp_depth:?hooks:print_hooklist->int->tCCFormat.printerendmoduletypeITER=sigtype'atvalto_iter:'at->'aIter.tvalof_iter:?init:'at->'aIter.t->'atvalto_list:'at->'alistvalof_list:?init:'at->'alist->'atendmoduletypeMONOID=sigtypetvalzero:tvalplus:t->t->tendmoduletypeGROUP=sigincludeMONOIDvalinverse:t->tend